import time
import pdb
import sys
import Orange
import orange
import heapq
sys.path.extend(['.', '..'])

from itertools import chain
from collections import deque


from ..learners.cn2sd.rule import fill_in_rules
from ..learners.cn2sd.refiner import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *
from ..util import *

from basic import Basic

inf = float('inf')
_logger = get_logger()

class Naive(Basic):
    def __init__(self, *args, **kwargs):
        Basic.__init__(self, *args, **kwargs)
        self.start = time.time()
        self.max_wait = kwargs.get('max_wait', 60*60*2) # 2 hours default
        self.n_rules_checked = 0
        self.stop = False
        self.cs = kwargs.get('cs', None)

        self.checkpoints_per_c = defaultdict(list)


    def __call__(self, full_table, bad_tables, good_tables, **kwargs):
        """
        table has been trimmed of extraneous columns.
        """
        self.setup_tables(full_table, bad_tables, good_tables, **kwargs)
        if not self.cs:
            self.cs = [self.c]
        
        self.bests_per_c = defaultdict(list)
        for c in self.cs:
            self.bests_per_c[c] = list()
            self.checkpoints_per_c[c] = list()

        self.max_complexity = kwargs.get('max_complexity', self.max_complexity)
        self.granularity = kwargs.get('granularity', self.granularity)
        discretes = [attr for attr in full_table.domain 
                     if attr.name in self.cols and attr.var_type == Orange.feature.Type.Discrete]


        self.all_clauses = map(self.all_unit_clauses, self.cols)
        self.col_to_clauses = dict(zip(self.cols, self.all_clauses))
        base_rule = SDRule(self.full_table, None)            


        start = time.time()
        if discretes:
            max_card = max([len(attr.values) for attr in discretes])
            for m in xrange(1, max_card+1):
                if self.stop:
                    break
                self.foo(base_rule, max_card=m)
        else:
            self.foo(base_rule)
        self.cost = time.time() - start


        # given a rule = c1,..,cn
        # options for each clause c_i
        # 1) extend c_i to the left, to the right
        # 2) append a new clause
        rules = self.bests_per_c[self.cs[0]]
        rules.sort(key=lambda r: r.quality, reverse=True)
        _logger.debug("best\n%s", "\n".join(map(lambda r: '%.4f\t%s' % (r.quality, str(r)), rules)))

        fill_in_rules(rules, full_table, cols=self.cols)
        clusters = [Cluster.from_rule(rule, self.cols, rule.quality) for rule in rules]
        self.all_clusters = clusters

        self.costs = {'cost' : self.cost}
        return clusters

    def max_card_in_conds(self, conds):
        lens = []
        for c in conds:
            attr = self.full_table.domain[c.position]
            if attr.var_type == Orange.feature.Type.Discrete:
                lens.append(len(attr.values))
        return lens and max(lens) or 0

    def foo(self, rule, max_card=None):
        for cols in powerset(self.cols):
            if not cols:
                continue
            if len(cols) > self.max_complexity:
                continue

            _logger.debug(str( cols))
            all_clauses = [self.get_all_clauses(col, max_card) for col in cols]
            for clauses in self.dfs(*all_clauses):
                new_rule = SDRule(rule.data, None, clauses, None)
                
                self.n_rules_checked -= len(clauses)
                if self.n_rules_checked <= 0:
                    diff = time.time() - self.start
                    for c in self.cs:
                        if (not self.checkpoints_per_c or 
                            not self.checkpoints_per_c[c] or 
                            diff - self.checkpoints_per_c[c][-1][0] > 10):
                            bests = self.bests_per_c[c]
                            if bests:
                                best_rule = max(bests, key=lambda r: r.quality)
                                clone = best_rule.clone()
                                clone.quality = best_rule.quality
                                self.checkpoints_per_c[c].append((diff, clone))
                    self.stop = diff > self.max_wait
                    self.n_rules_checked = 100

                    _logger.debug( "%.4f\t%d", time.time() - self.start, self.n_rules_checked)
                    _logger.debug(str( new_rule))

                if self.stop:
                    return

                if max_card is not None and self.max_card_in_conds(clauses) < max_card:
                    continue

                influences = [self.influence(new_rule, c) for c in self.cs]
                for c, influence in zip(self.cs, influences):
                    clone = new_rule.clone()
                    clone.quality = influence
                    clone.__examples__ = None

                    if not valid_number(clone.quality):
                        continue

                    if len(self.bests_per_c[c]) < self.max_bests:
                        heapq.heappush(self.bests_per_c[c], clone)
                    else:
                        heapq.heapreplace(self.bests_per_c[c], clone)
                new_rule.__examples__ = None

    def dfs(self, *iterables, **kwargs):
        q = kwargs.get('q', deque())
        if not iterables:
            yield q
            return

        for o in iterables[0]:
            q.append(o)
            for clauses in self.dfs(*iterables[1:], q=q):
                yield clauses
            q.pop()





import json
import math
import pdb
import random
import numpy as np
import sys
import time
sys.path.extend(['.', '..'])

from itertools import chain, repeat, izip
from collections import defaultdict
from operator import mul, and_, or_
from scipy.optimize import fsolve
from matplotlib import pyplot as plt

from ..util import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *

from frontier import *
from rangemerger import RangeMerger2

_logger = get_logger()


    
 
class StreamRangeMerger(RangeMerger2):
  def __init__(self, *args, **kwargs):
    super(StreamRangeMerger, self).__init__(*args, **kwargs)

    self.valid_cluster_f = kwargs.get('valid_cluster_f', lambda c: True)

    # idx -> clusters to expand  -- different than clusters on frontier!!
    self.tasks = defaultdict(list)

    # all values for each dimension
    self.all_cont_vals = defaultdict(set) # idx -> values
    # name -> { val -> [sum, count] }
    self.all_disc_vals = defaultdict(lambda: defaultdict(lambda: [0,0])) 

    # name -> { val -> # times failed }
    self.failed_disc_vals = defaultdict(lambda: defaultdict(lambda:0))

    # stores the frontier after each iteration
    self.added = set()
    self.seen = set()
    self.frontiers = []
    self.adj_graph = None


    self.K = 1
    self.nblocks = 50

    if len(self.learner.full_table) < 40000:
      self.K = 2
      self.nblocks = 60
    if len(self.learner.full_table) < 10000:
      self.nblocks = 100

    self.get_frontier = CheapFrontier(self.c_range, K=self.K, nblocks=self.nblocks)
    self.get_frontier.stats = self.stats


    if self.DEBUG:
      self.renderer = InfRenderer('/tmp/merger.pdf', c_range=self.c_range)

  def close(self):
    if self.DEBUG:
      self.renderer.close()

  def get_frontier_obj(self, version):
    while version >= len(self.frontiers):
      frontier = CheapFrontier(self.c_range, K=self.K, nblocks=self.blocks)
      frontier.stats = self.stats
      self.frontiers.append(frontier)
    return self.frontiers[version]

  @property
  def frontier_iter(self):
    return list(self.frontiers)


  @instrument
  def setup_stats(self, clusters):
    all_inf = lambda l: all([abs(v) == float('inf') for v in l])
    clusters = filter(lambda c: c.bound_hash not in self.added, clusters)
    clusters = filter(lambda c: not all_inf(c.inf_state[0]), clusters)
    clusters = filter(lambda c: len(c.inf_state[2]) == 0 or not all_inf(c.inf_state[2]), clusters)
    self.added.update([c.bound_hash for c in clusters])

    super(StreamRangeMerger, self).setup_stats(clusters)

    start = time.time()
    if not self.adj_graph:
      self.adj_graph = self.make_adjacency([], True)
    self.adj_graph.insert(clusters)
    self.adj_graph.sync()
    self.stats['adj_sync'][0] += time.time() - start
    self.stats['adj_sync'][1] += 1

    for c in clusters:
      for idx in xrange(len(c.cols)):
        self.all_cont_vals[idx].add(c.bbox[0][idx])
        self.all_cont_vals[idx].add(c.bbox[1][idx])
      for disc, vals in c.discretes.iteritems():
        if len(vals) < 3:
          vals = [(v,) for v in vals]
        else:
          vals = [tuple(vals)]
        for v in vals:
          self.all_disc_vals[disc][v][0] += c.inf_func(0.1)
          self.all_disc_vals[disc][v][1] += 1
        #self.all_disc_vals[disc].update(vals)

    return clusters

  @instrument
  def best_so_far(self, prune=False):
    clusters = set()
    for frontier in self.frontier_iter:
      clusters.update(frontier.frontier)

    if prune:
      for c in clusters:
        c.c_range = list(self.c_range)
      clusters = self.get_frontier(clusters)[0]
      clusters = filter(lambda c: r_vol(c.c_range), clusters)

    if self.DEBUG:
      self.renderer.new_page()
      self.renderer.set_title('best so far')
      self.renderer.plot_active_inf_curves(clusters)
    return clusters

  @instrument
  def best_at_c(self, c_val, K=6):
    clusters = set()
    for frontier in self.frontier_iter:
      clusters.update(frontier.seen_clusters)

    clusters = sorted(clusters, key=lambda c: c.inf_func(c_val), reverse=True)[:K]
    return clusters

  @instrument
  def add_clusters(self, clusters, idx=0):
    """
    Return list of new clusters that are on the frontier
    """
    if not clusters: return []

    if self.DEBUG:
      print "add_clusters"
      self.print_clusters(clusters)
      self.renderer.new_page()
      self.renderer.set_title("add_clusters %d clusters" % len(clusters))
      for f in self.frontier_iter:
        self.renderer.plot_inf_curves(f.frontier, color='grey') 
      self.renderer.plot_inf_curves(clusters, color='green')

    clusters = self.setup_stats(clusters)
    base_frontier = self.get_frontier_obj(idx)
    clusters, _ = base_frontier.update(clusters)

    if self.DEBUG:
      print "base_frontier"
      self.print_clusters(clusters)
      self.renderer.plot_active_inf_curves(clusters, color='red')

    # clear out current tasks
    self.tasks[idx] = filter(base_frontier.__contains__, self.tasks[idx])
    self.tasks[idx].extend(clusters)

    # remove non-frontier-based expansions from future expansion
    for tidx in self.tasks.keys():
      if tidx <= idx: continue
      checker = lambda c: not any(map(base_frontier.__contains__, c.ancestors))
      self.tasks[tidx] = filter(checker, self.tasks[tidx])

    if clusters:
      _logger.debug("merger:\tadded %d clusters\t%d tasks left", len(clusters), self.ntasks)
    return clusters


  @property
  def ntasks(self):
    if len(self.tasks) == 0: return 0
    return sum(map(len, self.tasks.values()))

  def has_next_task(self):
    if not self.tasks: return False
    return self.ntasks > 0

  def next_tasks(self, n=1):
    ret = []
    for tkey in reversed(self.tasks.keys()):
      tasks = self.tasks[tkey]
      while len(ret) < n and tasks:
        idx = random.randint(0, len(tasks)-1)
        ret.append((idx, tasks.pop(idx)))
    return ret



  @instrument
  def __call__(self, n=2):
    """
    Return any successfully expanded clusters (improvements)
    """
    nmerged = self.nmerged
    start = time.time()
    tasks = self.next_tasks(n)
    improvements = set()

    for idx, cluster in tasks:
      cur_frontier = self.get_frontier_obj(idx)
      next_frontier = self.get_frontier_obj(idx+1)
      new_clusters = self.run_task(idx, cluster, cur_frontier, next_frontier)

      debug = self.DEBUG
      self.DEBUG = False
      self.add_clusters(new_clusters, idx+1)
      self.DEBUG = debug
      improvements.update(new_clusters)

    _logger.debug("merger\ttook %.1f sec\t%d improved\t%d tried\t%d tasks left", 
        time.time()-start, len(improvements), (self.nmerged-nmerged), self.ntasks)
    return improvements


  def run_task(self, idx, cluster, cur_frontier, next_frontier):
    if not (idx == 0 or self.valid_cluster_f(cluster)):
      _logger.debug("merger\tbelow thresh skipping\t %s" % cluster)
      return []

    if self.DEBUG:
      self.renderer.new_page()
      self.renderer.set_title("expand %s" % str(cluster.rule))
      self.renderer.plot_inf_curves([cluster], color='grey') 

    self.rejected_disc_vals = defaultdict(list)
    self.rejected_cont_vals = defaultdict(set)
    expanded = self.greedy_expansion(cluster, self.seen, idx, cur_frontier)
    expanded = [c for c in expanded if c.bound_hash != cluster.bound_hash]

    if self.DEBUG:
      self.renderer.plot_inf_curves(expanded, color='green')

    cur_expanded, rms = cur_frontier.update(expanded)
    next_expanded, rms2 = next_frontier.update(cur_expanded)
    f = lambda c: c.bound_hash != cluster.bound_hash
    improved_clusters = set(filter(f, next_expanded))

    to_hash = lambda cs: set([c.bound_hash for c in cs])
    exp_bounds = to_hash(expanded)
    cur_bounds = to_hash(cur_expanded)
    next_bounds = to_hash(next_expanded)

    for c in chain(cur_expanded, rms):
      _logger.debug("merger\texpanded\tcur_idx(%s)\tnext_idx(%s)\t%.3f-%.3f\t%s", 
          (c.bound_hash in exp_bounds), 
          (c.bound_hash in next_bounds), 
          c.c_range[0], c.c_range[1],
          c.rule.simplify())
    if self.DEBUG:
      self.renderer.plot_active_inf_curves(cur_frontier.frontier, color='blue')
      self.renderer.plot_active_inf_curves(next_frontier.frontier, color='red')

    return improved_clusters


  @instrument
  def dims_to_expand(self, cluster, seen, version=None):
    for idx in xrange(len(cluster.cols)):
      vals = np.array(list(self.all_cont_vals[idx]))
      smaller = vals[(vals < cluster.bbox[0][idx])]
      bigger =  vals[(vals > cluster.bbox[1][idx])]
      yield idx, 'dec', smaller.tolist()
      yield idx, 'inc', bigger.tolist()

    for name, vals in cluster.discretes.iteritems():
      ret = []
      maxval = (len(vals) > 1) and max(vals) or None
      vals2infs = self.all_disc_vals[name].items()
      vals2infs.sort(key=lambda p: p[1][0] / float(p[1][1]+1.), reverse=True)

      for disc_vals, score in vals2infs:
        subset = set(disc_vals).difference(vals)
        subset.difference_update([v for v in subset if self.failed_disc_vals[name][str(v)] > 1])
        if maxval:
          subset = set(filter(lambda v: v >= maxval, subset))
        ret.append(subset)
      ret = filter(bool, ret)

      if ret:
        yield name, 'disc', ret
        return
        p = np.arange(len(ret), 0, -1).astype(float)
        p /= p.sum()
        ret = np.random.choice(ret, min(len(ret), 10), p=p, replace=False)
        yield name, 'disc', ret

  @instrument
  def check_direction(self, cluster, dim, direction, vals):
    key = cluster.bound_hash
    if direction == 'disc':
      for subset in self.rejected_disc_vals[dim]:
        if subset.issubset(vals): 
          return []
    if direction == 'inc':
      cont_vals = self.rejected_cont_vals[(dim, direction)]
      if cont_vals:
        vals = filter(lambda v: v > max(cont_vals), vals)
    if direction == 'dec':
      cont_vals = self.rejected_cont_vals[(dim, direction)]
      if cont_vals:
        vals = filter(lambda v: v < min(cont_vals), vals)
    return vals

  @instrument
  def update_rejected_directions(self, cluster, dim, direction, val):
    if direction == 'disc':
      for v in list(val):
        self.rejected_disc_vals[dim].append(set([v]))
        self.failed_disc_vals[dim][str(v)] += 1
    if direction == 'inc':
      self.rejected_cont_vals[(dim, direction)].add(round(val, 1))
    if direction == 'dec':
      self.rejected_cont_vals[(dim, direction)].add(round(val, 1))

  @instrument
  def greedy_expansion(self, cluster, seen, version=None, frontier=None):
    _logger.debug("merger\tgreedy_expand\t%s", cluster.rule.simplify())
    if frontier is None:
      frontier = CheapFrontier(self.c_range, K=1, nblocks=15)
      frontier.stats = self.stats
      frontier.update([cluster])

    cols = cluster.cols
    for dim, direction, vals in self.dims_to_expand(cluster, seen, version=version):
      if len(vals) == 0: continue
      attrname = isinstance(dim, basestring) and dim or cols[dim]
      vals = self.check_direction(cluster, dim, direction, vals)
      realvals = self.pick_expansion_vals(cluster, dim, direction, vals)

      nfails = 0
      for v in realvals:
        tmp = None
        if direction == 'inc':
          tmp = self.dim_merge(cluster, dim, None, v, seen)
        elif direction == 'dec':
          tmp = self.dim_merge(cluster, dim, v, None, seen)
        else:
          tmp = self.disc_merge(cluster, dim, v)

        if not tmp: 
          _logger.debug("merger\tnoexpand\t%s\t%s\t%s options", attrname[:15], direction, len(vals))
          continue

        improvements = frontier.improvement(tmp)
        if improvements.max() > 0:
          print str(tmp)
          print "\t", [round(v,2) for v in improvements]
        frontier.update([tmp])
        isbetter = tmp in frontier
        _logger.debug("merger\tcand\t%s\t%s\t%s\t%s", attrname[:15], direction, isbetter, v)
 
        seen.add(tmp.bound_hash)
        if not isbetter:
          self.update_rejected_directions(cluster, dim, direction, v)
          if direction != 'disc':
            nfails += 1
            if nfails > 1:
              break
        if direction != 'disc':
          cluster = tmp

    return frontier.frontier


class PartitionedStreamRangeMerger(StreamRangeMerger):

  def __init__(self, *args, **kwargs):
    super(PartitionedStreamRangeMerger, self).__init__(*args, **kwargs)
    self.frontiers = defaultdict(list)
    self.tasks = defaultdict(list)

  def get_frontier_obj(self, version, partitionkey):
    frontiers = self.frontiers[partitionkey]
    while version >= len(frontiers):
      frontier = CheapFrontier(self.c_range, K=self.K, nblocks=self.nblocks)
      frontier.stats = self.stats
      frontiers.append(frontier)
    return frontiers[version]
  
  @property
  def frontier_iter(self):
    return chain(*self.frontiers.values())

  @instrument
  def add_clusters(self, 
      clusters, idx=0, partitionkey=None, skip_frontier=False):
    """
    Return list of new clusters that are on the frontier
    """
    if partitionkey is None:
      raise RuntimeError('addclusters partitionkey cannot be none')

    if not clusters: return []

    print "add %d clusters" % len(clusters)
    if self.DEBUG:
      self.renderer.new_page()
      self.renderer.set_title("add_clusters %d clusters" % len(clusters))
      for f in self.frontier_iter:
        self.renderer.plot_inf_curves(f.frontier, color='grey') 
      self.renderer.plot_inf_curves(clusters, color='green')

    nclusters = len(clusters)
    clusters = self.setup_stats(clusters)
    frontier = self.get_frontier_obj(idx, partitionkey)
    if not skip_frontier:
      clusters, _ = frontier.update(clusters)

    if self.DEBUG and not skip_frontier:
      print "base_frontier"
      self.print_clusters(clusters)

    if self.DEBUG:
      self.renderer.plot_active_inf_curves(clusters, color='red')

    # clear out current tasks
    tkey = (partitionkey, idx)
    self.tasks[tkey] = filter(frontier.__contains__, self.tasks[tkey])
    self.tasks[tkey].extend(clusters)

    # remove non-frontier-based expansions from future expansion
    for (pkey, tidx) in self.tasks.keys():
      if pkey != partitionkey: continue
      if tidx <= idx: continue
      checker = lambda c: not any(map(frontier.__contains__, c.ancestors))
      self.tasks[tkey] = filter(checker, self.tasks[tkey])

    _logger.debug("merger\t%s\tadd %d of %d clusters\t%d idx\t%d tasks left", partitionkey, len(clusters), nclusters, idx, self.ntasks)
    return clusters


  def next_tasks(self, n=1):
    ret = []
    for tkey in reversed(self.tasks.keys()):
      if len(ret) >= n: break
      tasks = self.tasks[tkey]
      ntasks = len(tasks)
      if not ntasks: continue
      idxs = np.random.choice(ntasks, min(ntasks, n-len(ret)), replace=False).tolist()
      for idx in sorted(idxs, reverse=True):
        ret.append((tkey[0], tkey[1], tasks.pop(idx)))
    return ret


  def __call__(self, n=2):
    nmerged = self.nmerged
    start = time.time()
    tasks = self.next_tasks(n)
    improvements = set()

    for pkey, idx, cluster in tasks:
      cur_frontier = self.get_frontier_obj(idx, pkey)
      next_frontier = self.get_frontier_obj(idx+1, pkey)
      new_clusters = self.run_task(idx, cluster, cur_frontier, next_frontier)

      self.add_clusters(new_clusters, idx=idx+1, partitionkey=pkey, skip_frontier=True)
      improvements.update(new_clusters)

      _logger.debug("merger\t%s\ttook %.1f sec\t%d improved\t%d tried\t%d tasks left", 
          pkey, time.time()-start, len(improvements), (self.nmerged-nmerged), self.ntasks)
    return improvements







import sys
import pdb
import random
sys.path.extend(('.', '..'))


from common import *
from util import get_logger

_logger = get_logger()
random.seed(0)


# configurations
# ndt: nothing, c
# basic: max complexity, c
# bdt: epsilon, tau, lambda, c
# mc: c
def print_stats(pp, stats, rules, title):
  fig = plt.figure(figsize=(16, 4))
  plt.suptitle(title)
  ss = [rule.quality for rule in rules]
  accs, ps, rs, f1 = zip(*stats)
  for idx, (ys, statname) in enumerate(((accs, 'acc'), (ps, 'prec'), (rs, 'recall'), (f1, 'f1'))):
    sub = fig.add_subplot(int('14'+str(idx)))
    sub.scatter(ss, ys, lw=0)

    sub.set_ylim(0, 1.2)
    sub.set_title(statname)

  plt.savefig(pp, format='pdf')


def make_params(**kwargs):
  ret = []
  for klass in kwargs.get('klasses', [NDT, BDT, MR, Naive]):
    for c in kwargs.get('c', [0, 0.5, 0.75, 1.]):

      params = dict(kwargs)
      params.update({
        'c':c, 
        'klass':klass, 
        'klassname' : klass.__name__[-7:], 
        'p' : 0.6,
        'cols' : kwargs.get('cols', None)
      })

      if klass == BDT:
        for epsilon in kwargs.get('epsilon', [0.001, 0.01, 0.05]):
          for l in kwargs.get('l', [0.5, 1.]):
            params2 = dict(params)
            params2.update({ 
              'epsilon':epsilon, 
              'l':l, 
              'min_improvement' : .01})
            ret.append(params2)
      elif klass == Naive:
        for l in kwargs.get('l', [0.5]):
          params2 = dict(params)
          params2['max_complexity'] = kwargs.get('max_complexity',4)
          params2['granularity'] = kwargs.get('granularity', 20) 
          params2['l'] = l
          ret.append(params2)
      else:
        for l in kwargs.get('l', [0.5, 1.]):
          params['l'] = l
          ret.append(params)

  ret.sort(key=lambda p: (p['klassname'], p['c']))
  return ret

def mkfmt(arr):
  """
  Given a list of object types, returns a tab separated formatting str
  """
  mapping = [(float, '%.4f'), (int, '%d'),  (object, '%s')]
  fmt = []
  for v in arr:
      for t,f in mapping:
          if isinstance(v,t):
              fmt.append(f)
              break

  return '\t'.join(fmt)


def get_ground_truth(db, datasetidx, c):
  """
  The experiments are supposed to also log the ground truth.
  Retriev them from the experiment stats table
  """
  with db.begin() as conn:
    q = """select ids from stats where klass = 'Naive' and dataset = %s and c = %s order by expid desc"""
    for row in conn.execute(q, datasetidx, c).fetchall():
        return map(int, map(float, row[0].split(',')))
    return None



def run(db, datasetidx, pp=None, **kwargs):
  expid = nextexpid(db)
  for params in make_params(**kwargs):
    run_params(db, datasetidx, expid, pp=pp, **params)
  try:
    complete(db, expid)
  except:
    import traceback
    traceback.print_exc()



def run_params(db, datasetidx, expid, pp=None, **params):
  pkeys = ['klassname', 'cols', 'epsilon', 'c', 'lambda']
  try:
    run_id = str(params)
    costs, rules, all_ids, table_size, learner = run_experiment(datasetidx, **params)
    total_cost = costs['cost_total']
    if params['klassname'] == 'Naive':
      rules = rules[:1]
      all_ids = all_ids[:1]
      truth = all_ids[0]
    else:
      truth = get_ground_truth(db, datasetidx, params['c'])

    if truth == None:
      pdb.set_trace()
      raise RuntimeError("Could not find ground truth for %d, %.4f" % (datasetidx, params['c']))

    all_stats = [compute_stats(ids, truth, table_size) for ids in all_ids]

    row_id = None
    cs = params.get('c', params.get('cs', []))
    for c in cs:
      for (acc, p, r, f1), rule, ids in zip(all_stats, rules, all_ids):
        ids_str = ','.join(map(str, ids))
        isbest = rule.isbest
        vals = [expid, str(datasetidx), params.get('notes', '')]
        vals.extend([params.get(key, None) for key in pkeys])
        vals[-2] = c
        vals.extend((total_cost, acc, p, r, f1, rule.quality, isbest, str(rule), ids_str))
        row_id = save_result(db, vals, costs)
        print mkfmt(vals[:-1]) % tuple(vals[:-1])

      if params['klassname'] == 'Naive':
        rules = rules[:1]
        all_ids = all_ids[:1]
        truth = all_ids[0]
        if not isinstance(cs, list): cs = [cs]
        for c in cs:
          for timein, rule in learner.checkpoints_per_c[c]:
            chk_ids = get_row_ids(rule, learner.full_table)
            acc, p, r, f1 = compute_stats(chk_ids, truth, table_size)
            ids_str = ','.join(map(str, map(int, chk_ids)))
            vals = [expid, datasetidx, params.get('notes', '')]
            vals.extend([params.get(key, None) for key in pkeys])
            vals[-2] = c
            vals.extend((timein, acc, p, r, f1, rule.quality, False, str(rule), True, ids_str))
            row_id = save_checkpoint(db, vals)
            print mkfmt(vals[:-1]) % tuple(vals[:-1])

    if pp:
      pass
      #print_stats(pp, all_stats, rules, ','.join(map(lambda p: '%s:%s'%tuple(p), params.items())))

  except Exception as e:
    import traceback
    traceback.print_exc()
    _logger.error(traceback.format_exc())

def save_result(db, stat, costs):
  with db.begin() as conn:
    q = """insert into stats(expid, dataset, notes, klass, cols, epsilon, c,     
                              lambda, cost, acc, prec, recall, f1, score, 
                              isbest, rule, ids) values(%s) returning id""" % (','.join(['%s']*len(stat)))
    sid = conn.execute(q, *stat).fetchone()[0]

    q = """insert into costs(sid, name, cost) values(%s,%s,%s)"""
    for name, cost in costs.items():
        if isinstance(cost, list):
            cost = cost[0]
        conn.execute(q, sid, name, cost)
    return sid

def save_checkpoint(db, stat):
  with db.begin() as conn:
    q = """insert into stats(expid, dataset, notes, klass, cols, epsilon, c,     
                              lambda, cost, acc, prec, recall, f1, score, 
                              isbest, rule, ischeckpoint, ids) values(%s) returning id""" % (','.join(['%s']*len(stat)))
    sid = conn.execute(q, *stat).fetchone()[0]
    return sid



def complete(db, expid):
  with db.begin() as conn:
    q = """update stats set completed = TRUE where expid = %s"""
    conn.execute(q, expid)



def nextexpid(db):
  with db.begin() as conn:
    q = """select max(expid)+1 from stats"""
    row = conn.execute(q).fetchone()
    expid = row[0]
    if expid == None:
        return 0
    return expid








import time
import pdb
import sys
import Orange
import orange
import heapq
sys.path.extend(['.', '..'])

import numpy as np
from itertools import chain
from collections import defaultdict
from sklearn.cluster import AffinityPropagation, KMeans
from scorpionsql.errfunc import ErrTypes, compute_bad_inf, compute_bad_score, compute_influence


from ..learners.cn2sd.rule import fill_in_rules
from ..learners.cn2sd.refiner import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *
from ..util import *


INF = inf = float('inf') # slowly replace inf with INF...



class Basic(object):


  def __init__(self, **kwargs):
    self.obj = kwargs.get('obj', None)
    self.aggerr = kwargs.get('aggerr', None)
    self.cols = list(self.aggerr.agg.cols)
    self.err_func = kwargs.get('err_func', self.aggerr.error_func.clone())
    self.merger = None
    self.params = {}
    self.costs = {}
    self.stats = defaultdict(lambda: [0, 0])  # used by @instrument

    self.bad_thresh = 0
    self.good_thresh = 0
    self.min_pts = kwargs.get('min_pts', 5)
    self.max_bests = 20
    self.max_complexity = kwargs.get('max_complexity', 4)
    self.granularity = kwargs.get('granularity', 100)

    self.l = kwargs.get('l', 0.5)
    self.c = kwargs.get('c', 0.3)
    self.c_range = kwargs.get('c_range', [0.05, 0.5])
    self.epsilon = kwargs.get('epsilon', 0.0001)
    self.tau = kwargs.get('tau', [0.1, 0.5])
    self.p = kwargs.get('p', 0.5)
    self.bincremental = kwargs.get('bincremental', True)
    self.use_cache = kwargs.get('use_cache', False)
    self.parallel = kwargs.get('parallel', False)

    self.DEBUG = kwargs.get('DEBUG', False)

    self.tablename = kwargs.get('tablename', None)


    self.scorer_cost = 0.
    self.merge_cost = 0.

    self.update_status = self.obj.update_status
    self.update_rules = self.obj.update_rules

    self.set_params(**kwargs)

  def __hash__(self):
    components = [
            self.__class__.__name__,
            str(self.aggerr.__class__.__name__),
            str(set(self.cols)),
            self.err_func.__class__.__name__,
            self.tablename,
            self.l,
            self.c
            ]
    components = map(str, components)
    return hash('\n'.join(components))

  def merge_stats(self, stats, prefix=''):
    for key, stat in stats.iteritems():
      mykey = '%s%s' % (prefix, key)
      if mykey in self.stats:
        self.stats[mykey][0] += stat[0]
        self.stats[mykey][1] += stat[1]
      else:
        self.stats[mykey] = list(stat)
              

  def set_params(self, **kwargs):
    self.cols = kwargs.get('cols', self.cols)
    self.use_cache = kwargs.get('use_cache', self.use_cache)
    self.params.update(kwargs)

  def setup_tables(self, full_table, bad_tables, good_tables, **kwargs):
    """
    create bad_err_funcs
    """
    self.full_table = full_table
    self.dummy_table = Orange.data.Table(full_table.domain)
    self.bad_tables = bad_tables
    self.good_tables = good_tables
    
    self.bad_err_funcs = self.aggerr.bad_error_funcs()
    self.good_err_funcs = self.aggerr.good_error_funcs(good_tables)

    for ef, t in zip(self.bad_err_funcs, bad_tables):
      ef.setup(t)

    for ef, t in zip(self.good_err_funcs, good_tables):
      ef.setup(t)

    domain = self.full_table.domain
    attrnames = [attr.name for attr in domain]
    self.cont_dists = dict(zip(attrnames, Orange.statistics.basic.Domain(self.full_table)))
    self.disc_dists = dict(zip(attrnames, Orange.statistics.distribution.Domain(self.full_table)))

  def __call__(self, full_table, bad_tables, good_tables, **kwargs):
    """
    table has been trimmed of extraneous columns.
    @return final_clusters
    """

    self.setup_tables(full_table, bad_tables, good_tables, **kwargs)

    pass

  def rule_complexity(self, rule):
    ret = 0
    for cond in rule.filter.conditions:
      pos = cond.position
      attr = self.full_table.domain[pos]
      if attr.var_type == Orange.feature.Type.Discrete:
        fd = self.disc_dists[attr.name]
        if len(cond.values) != len(fd.values()):
          ret += 1
          ret += max(0, 0.01 * len(cond.values) - 2)
      else:
        fb = self.cont_dists[attr.name]
        if not r_contains([cond.min, cond.max], [fb.min, fb.max]):
          ret += 1
    return ret


  def create_inf_func(self, cluster):
    """
    Args:
      cluster-like object with attributes:
        .inf_state
        .rule

    """
    inf_state = cluster.inf_state
    if inf_state is None:
      raise Exception("inf_state is None, cant' create inf_func")

    l = self.l
    vs = [abs(gv) for gv, gc in zip(inf_state[2], inf_state[3]) if gc]
    #maxg = max(vs) if vs else 0
    maxg = np.percentile(vs, 85) if vs else 0

    bds, bcs = [], []
    for idx in xrange(len(inf_state[0])):
      bd, bc = inf_state[0][idx], inf_state[1][idx]
      if valid_number(bd) and valid_number(bc):
        bds.append(bd)
        bcs.append(bc)
      else:
        bds.append(0)
        bcs.append(0)

    rule = cluster.rule
    complexity = self.rule_complexity(rule)
    f = lambda c: compute_influence(l, compute_bad_score(bds, bcs, c), maxg, nclauses=complexity)
    #f = np.frompyfunc(f, 1, 1)
    return f



  @instrument
  def cluster_to_rule(self, cluster, table=None):
    if not table:
      table = self.dummy_table
    rule = cluster.to_rule(
        table, cont_dists=self.cont_dists, disc_dists=self.disc_dists
    )
    return rule


  @instrument
  def influence_state(self, rule):
    bdeltas, bcounts = self.compute_stat(rule, self.bad_err_funcs, self.bad_tables)
    gdeltas, gcounts = self.compute_stat(rule, self.good_err_funcs, self.good_tables)
    gdeltas = map(abs, gdeltas)
    return bdeltas, bcounts, gdeltas, gcounts

  @instrument
  def influence_from_state(self, bdeltas, bcounts, gdeltas, gcounts, c=None, nclauses=0):
    if c is None:
        c = self.c
    binf = compute_bad_score(bdeltas, bcounts, c)
    ginfs = [gdelta for gdelta,gcount in zip(gdeltas, gcounts) if gcount]
    ginf = ginfs and max(ginfs) or 0
    return compute_influence(self.l, binf, ginf, nclauses=nclauses)


  @instrument
  def influence_cluster(self, cluster, table=None):
    rule = self.cluster_to_rule(cluster, table)
    cluster.error = self.influence(rule)
    cluster.inf_state = rule.inf_state
    return cluster.error


  @instrument
  def influence(self, rule, c=None):
    inf_state = self.influence_state(rule)
    quality = self.influence_from_state(*inf_state, nclauses=len(rule.filter.conditions))
    rule.quality = quality
    rule.inf_state = inf_state
    return quality


  def compute_stat(self, rule, err_funcs, tables):
    datas = rule and map(rule.filter_table, tables) or tables
    infs = []
    lens = []
    for idx, (ef, data) in enumerate(zip(err_funcs, datas)):
      if len(data) == len(tables[idx]):
        influence = -INF
      else:
        arr = data.to_numpyMA('ac')[0]
        influence = ef(arr.data)
      infs.append(influence)
      lens.append(len(data))
    return infs, lens





  def all_unit_clauses(self, attr):
      # continuous: split 1000 ways, uniformly
      # discrete: every unique value
      attrobj = self.full_table.domain[attr]
      idx = self.full_table.domain.index(attrobj)
      if attrobj.var_type == Orange.feature.Type.Discrete:
          ddist = Orange.statistics.distribution.Domain(self.full_table)[idx]
          return ddist.keys()
      
      bdist = Orange.statistics.basic.Domain(self.full_table)[idx]
      minv, maxv = bdist.min, bdist.max
      if minv == maxv:
          return [[-INF, INF]]

      block = (maxv - minv) / self.granularity
      ranges = [[minv + i*block, minv + (i+1)*block] for i in xrange(self.granularity)]
      ranges[0][0] = -INF
      ranges[-1][1] = INF
      return ranges


  def get_all_clauses(self, attr, max_card):
      class Ret(object):
          def __init__(self, attr, max_card, par):
              self.attr = attr
              self.max_card = max_card
              self.par = par

          def __iter__(self):
              attrobj = self.par.full_table.domain[self.attr]
              if attrobj.var_type == Orange.feature.Type.Discrete:
                  return self.par.all_discrete_clauses(self.attr, self.max_card)
              else:
                  return self.par.all_cont_clauses(self.attr)
      return Ret(attr, max_card, self)

          
  def all_discrete_clauses(self, attr, max_card=None):
      all_vals = self.col_to_clauses[attr]
      attrobj = self.full_table.domain[attr]
      idx = self.full_table.domain.index(attrobj)
      
      if max_card:
          for card in xrange(1, max_card+1):
              for vals in combinations(all_vals, card):
                  vals = [orange.Value(attrobj, value) for value in vals]
                  yield orange.ValueFilter_discrete(
                          position = idx,
                          values = vals)
      else:
          for vals in powerset(all_vals):
              vals = [orange.Value(attrobj, value) for value in vals]
              yield orange.ValueFilter_discrete(
                      position = idx,
                      values = vals)


  def all_cont_clauses(self, attr):
      units = self.col_to_clauses[attr]
      idx = self.full_table.domain.index(self.full_table.domain[attr])
      for sidx in xrange(0, len(units)):
          for eidx in xrange(sidx, len(units)):
              minv = units[sidx][0]
              maxv = units[eidx][1]
              yield Orange.data.filter.ValueFilterContinuous(
                      position=idx,
                      oper=orange.ValueFilter.Between,
                      min=minv,
                      max=maxv)





import math
import pdb
import random
import numpy as np
import sys
import time
import Orange
sys.path.extend(['.', '..'])

from itertools import chain, repeat
from collections import defaultdict
from rtree.index import Index as RTree
from rtree.index import Property as RProp
from operator import mul, and_, or_
from sklearn.neighbors import NearestNeighbors, DistanceMetric

from ..util import rm_attr_from_domain, get_logger
from ..util.table import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *

_logger = get_logger()

class FeatureMapper(object):
  """
  For discrete features
  """

  def __init__(self, domain, cont_dists):
    self.feature_mappers = {}
    self.cont_dists = cont_dists
    self.ranges = { col: cont_dists[col].max - cont_dists[col].min for col in cont_dists.keys() if cont_dists[col] }
    for attr in domain:
      if attr.var_type == Orange.feature.Type.Discrete:
        self.feature_mappers[attr.name] = self.get_feature_mapper(attr)

  def get_feature_mapper(self, attr):
    name = attr.name
    val2idx = {v:idx for idx,v in enumerate(attr.values)}
    return val2idx

  def __attrs__(self):
    return self.feature_mappers.keys()
  attrs = property(__attrs__)

  def nvals(self, name):
    return len(self.feature_mappers.get(name, []))

  def __call__(self, cluster, name):
    vals = cluster.discretes.get(name, None)
    mapping = self.feature_mappers.get(name, {})
    if vals is None: 
      return np.ones(len(mapping))

    vect = np.zeros(len(mapping)).astype(int)
    for v in vals:
      vect[v] = 1
    return vect

class AdjacencyGraph(object):
  """
  Stores versions of adjacency graphs and manages insert buffering
  """
  def __init__(self, clusters, domain, cont_dists):
    """
    Args
      domain: orange table domain
    """
    self.feature_mapper = FeatureMapper(domain, cont_dists)
    self.versions = []
    self.insert_bufs = defaultdict(set)

    if clusters:
      self.insert(clusters)
      self.sync()

  def sync(self):
    """
    apply all of the pending inserts and deletes to the existing
    versions
    """
    for idx, v in enumerate(self.versions):
      buf = self.insert_bufs[idx]
      if not buf: continue
      buf.update(v.clusters)
      self.versions[idx] = AdjacencyVersion(self.feature_mapper)
      self.versions[idx].bulk_init(list(buf))
      self.insert_bufs[idx] = set()

  def new_version(self, clusters=None):
    clusters = clusters or []
    v = AdjacencyVersion(self.feature_mapper)
    self.versions.append(v)

  def ensure_version(self, version):
    if version is None: return
    while len(self.versions) <= version:
      self.new_version()
    return self.versions[version]

  def __len__(self):
    return len(self.versions)

  def insert(self, clusters, version=0):
    self.ensure_version(version)
    if not isinstance(clusters, list):
      clusters = list(clusters)

    v = self.versions[version]
    clusters = [c for c in clusters if not v.contains(c)]
    self.insert_bufs[version].update(clusters)
    return len(clusters)

  def remove(self, clusters, version=0):
    self.ensure_version(version)
    if not isinstance(clusters, list):
      clusters = list(clusters)

    v = self.versions[version]
    rms = [v.remove(cluster) for cluster in clusters]
    return rms

  def neighbors(self, cluster, version=None):
    self.ensure_version(version)
    if version is not None:
      return self.versions[version].neighbors(cluster)

    ret = []
    for v in self.versions:
      ret.extend(v.neighbors(cluster))
    return ret


class AdjacencyVersion(object):

  def __init__(self, feature_mapper):
    #self.partitions_complete = partitions_complete
    self.cid = 0
    self.disc_idxs = {}
    self.feature_mapper = feature_mapper
    self.radius = .15
    self.metric = 'hamming'

    self._rtree = None  # internal datastructure
    self._ndim = None
    self.clusters = []
    self.id2c = dict()
    self.c2id = dict()

  def to_json(self):
    data = {
            'clusters' : [c and c.__dict__ or None for c in self.clusters],
            'id2c' : [(key, c.__dict__) for key, c in self.id2c.items()],
            'c2id' : [(c.__dict__, val) for c, val in self.c2id.items()],
            'cid' : self.cid,
            '_ndim' : self._ndim,
            '_rtreename' : 'BLAH'
            }
    return json.dumps(data)

  def from_json(self, encoded):
    data = json.loads(encoded)
    self.clusters = [c and Cluster.from_dict(c) or None for c in data['clusters']]
    self.id2c = dict([(key, Cluster.from_dict(val)) for key, val in data['id2c']])
    self.c2id = dict([(Cluster.from_dict(key), val) for key, val in data['c2id']])
    self.cid = data['cid']
    self._ndim = data['_ndim']
    self._rtree = None

  def setup_rtree(self, ndim, clusters=None):
    if self._rtree:
        return self._rtree

    self._ndim = ndim
    if not ndim:
        class k(object):
            def __init__(self, graph):
                self.graph = graph
            def insert(self, *args, **kwargs):
                pass
            def delete(self, *args, **kwargs):
                pass
            def intersection(self, *args, **kwargs):
                return xrange(len(self.graph.clusters))
        self._rtree = k(self)
        return self._rtree


    p = RProp()
    p.dimension = max(2, ndim)
    p.dat_extension = 'data'
    p.idx_extension = 'index'

    if clusters:
        gen_func = ((i, self.bbox_rtree(c, enlarge=0.005), None) for i, c in enumerate(clusters))
        self._rtree = RTree(gen_func, properties=p)
    else:
        self._rtree = RTree(properties=p)
    return self._rtree

  def bbox_rtree(self, cluster, enlarge=0.):
    cols = cluster.cols
    bbox = cluster.bbox
    lower, higher = map(list, bbox)
    if self._ndim == 1:
      lower.append(0)
      higher.append(1)

    if enlarge != 0:
      for idx, col in enumerate(cols):
        rng = enlarge * self.feature_mapper.ranges[col]
        lower[idx] -= rng
        higher[idx] += rng

    bbox = lower + higher
    return bbox

  def insert_rtree(self, idx, cluster):
    self.setup_rtree(len(cluster.bbox[0]))
    self._rtree.insert(idx,self.bbox_rtree(cluster))
    return cluster

  def remove_rtree(self, idx, cluster):
    self.setup_rtree(len(cluster.bbox[0]))
    self._rtree.delete(idx, self.bbox_rtree(cluster))
    return cluster

  def search_rtree(self, cluster):
    self.setup_rtree(len(cluster.bbox[0]))
    bbox = self.bbox_rtree(cluster, enlarge=0.01)
    return self._rtree.intersection(bbox)
    res = [self.clusters[idx] for idx in self._rtree.intersection(bbox)]
    return filter(bool, res)

  def bulk_init(self, clusters):
    if not clusters: return

    self.setup_rtree(len(clusters[0].bbox[0]), clusters)
    self.clusters = clusters
    for cid, c in enumerate(clusters):
      self.id2c[cid] = c
      self.c2id[c] = cid
    
    for dim in self.feature_mapper.attrs:
      Xs = []
      for cidx, c in enumerate(clusters):
        Xs.append(self.feature_mapper(c, dim))
      idx = NearestNeighbors(
          radius=self.radius, 
          algorithm='ball_tree', 
          metric=self.metric
      )
      self.disc_idxs[dim] = idx
      self.disc_idxs[dim].fit(np.array(Xs))

  def contains(self, cluster):
    return cluster in self.c2id
  
  def remove(self, cluster):
    if cluster in self.c2id:
      cid = self.c2id[cluster]
      self.remove_rtree(cid, cluster)
      del self.c2id[cluster]
      del self.id2c[cid]
      self.clusters[cid] = None
      return True
    return False


  def neighbors(self, cluster):
    ret = None
    for name, vals in cluster.discretes.iteritems():
      if name not in self.disc_idxs:
        return []
      vect = self.feature_mapper(cluster, name)
      index = self.disc_idxs[name]
      dists, idxs = index.radius_neighbors(vect, radius=self.radius)
      idxs = set(idxs[0].tolist())

      if ret is None:
        ret = idxs
      else:
        ret.intersection_update(idxs)
        #ret.update(idxs)
      if not ret: return []

    idxs = self.search_rtree(cluster)
    if ret is None:
      ret = set(idxs)
    else:
      ret.intersection_update(set(idxs))

    return filter(bool, [self.clusters[idx] for idx in ret])


  """
  def neighbors(self, cluster):
    if not self.partitions_complete:
        return filter(bool, self.clusters)

    if cluster in self.graph:
        return self.graph[cluster]

    ret = set()
    intersects = self.search_rtree(cluster)
    for key in filter(cluster.adjacent, intersects):
        if box_completely_contained(key.bbox, cluster.bbox):
            continue
        ret.update(self.graph[key])
    return ret
    """







#
# helper methods for manipulating bounds
#

def r_vol(bound):
  return max(0, bound[1] - bound[0])

def r_empty(bound):
  return bound[1] <= bound[0]

def r_equal(bound1, bound2):
  return bound1[0] == bound2[0] and bound1[1] == bound2[1]

def r_lt(bound1, bound2):
  "bound1 values < bound2 values"
  return bound1[0] < bound2[0] and bound1[1] < bound2[1]

def r_lte(bound1, bound2):
  "bound1 values <= bound2 values"
  return bound1[0] <= bound2[0] and bound1[1] <= bound2[1]

def r_expand(bound, perc=0.05):
  v = bound[1] - bound[0]
  v *= perc
  return [bound[0]-v, bound[1]+v]

def r_scontains(bound1, bound2):
  "bound1 strictly contains bound2"
  return bound1[0] < bound2[0] and bound2[1] < bound1[1]

def r_contains(bound1, bound2):
  "bound1  contains bound2"
  return bound1[0] <= bound2[0] and bound2[1] <= bound1[1]


def r_intersect(bound1, bound2):
  return [max(bound1[0], bound2[0]), min(bound1[1], bound2[1])]

def r_union(bound1, bound2):
  return [min(bound1[0], bound2[0]), max(bound1[1], bound2[1])]

def r_subtract(bound1, bound2):
  """
  remove bound2 from bound1. 
  Return list of bound
  """
  if r_contains(bound2, bound1):
    return [ [bound1[0], bound1[0]] ]
  if r_scontains(bound1, bound2):
    return [ [bound1[0], bound2[0]], [bound2[1], bound1[1]] ]
  inter = r_intersect(bound1, bound2)
  if r_lte(inter, bound1):
    return [ [inter[1], bound1[1]] ]
  return [ [bound1[0], inter[0]] ]





#!/usr/bin/env python
try:
    from setuptools import setup, find_packages
except ImportError:
    import ez_setup
    ez_setup.use_setuptools()
from setuptools import setup, find_packages
import scorpion

setup(name="scorpion",
      version=scorpion.__version__,
      description="scorpion",
      license="MIT",
      author="Eugene Wu",
      author_email="eugenewu@mit.edu",
      url="http://github.com/sirrice/scorpion",
      include_package_data = True,      
      packages = find_packages(),
      package_dir = {'scorpion' : 'scorpion'},
      scripts = [
        'tests/unit/adjgraph.py',
        'tests/unit/sharedobj.py',
        'tests/unit/render.py',
        'tests/unit/sharedobj.py',
        'bin/fixnulls.py',
      ],
      package_data = {
        'scorpion': [
          'jars/*'
        ]
      },
      install_requires = [
        'flask', 'psycopg2', 'sqlalchemy', 
        'Orange', 'numpy', 'scipy', 'scikit-learn',
        'matplotlib', 'pyparsing', 'rtree',
        'click'
      ],
      keywords= "")



import random

from scorpion.bottomup.cluster import *
from scorpion.util.render import *


random.seed(0)

clusters = []
for i in xrange(50):
  x,y = random.randint(0, 90), random.randint(0, 90)
  bbox = [[x, y], [x+10, y+10]]
  c = Cluster(bbox, random.random(), ['x', 'y'])
  c.inf_state = [
      [random.randint(1, 50) for i in range(10)],
      [random.randint(1, 10) for i in range(10)],
      [],
      []
  ]
  c.c_range = [0, 1]
  c.inf_func = c.create_inf_func(0.5)
  clusters.append(c)

renderer = ClusterRenderer("/tmp/test.pdf")
renderer.plot_title("test title")
renderer.plot_clusters(clusters)
renderer.close()


renderer = InfRenderer('/tmp/infs.pdf')
renderer.plot_title("boooo")
renderer.plot_inf_curves(clusters)
renderer.close()



from common import *
from test import run

if __name__ == '__main__':
  from sqlalchemy import create_engine
  db = create_engine('postgresql://localhost/sigmod')
  pp = PdfPages('figs/test.pdf')

  print "did you run setup_experiments.py to generate the testing env?"
  run(db, 19, pp, klasses=[MR], max_wait=10*60, c=[0., .25, .5, .75, 1.], granularity=20)
  exit()

  # run and gather "ground truth" for everything
  run(db, 0, pp, klasses=[Naive], max_wait=10*60, c=[0., .25, .5, .75, 1.], granularity=20, naive=True)
  run(db, 11, pp, klasses=[Naive], max_wait=10*60, c=[0., .25, .5, .75, 1.], granularity=20, naive=True)
  run(db, 5, pp, klasses=[Naive], max_wait=10*60, c=[0., .25, .5, .75, 1.], granularity=20, naive=True)
  run(db, 15, pp, klasses=[Naive], max_wait=5*60, c=[0., .25, .5, .75, 1.], granularity=20, naive=True)


  # run all others on intel_noon
  run(db, 0, pp, klasses=[NDT], c=[0., 0.25, 0.5, 0.75, 1.])
  run(db, 0, pp, klasses= [BDT], l=[0.5], c=[0., 0.25, 0.5, 0.75, 1.], epsilon=[0.0001, 0.001], tau=[0.1, 0.45])
  run(db, 0, pp, klasses= [BDT], l=[0.5], c=[1.], epsilon=[0.01, 0.1], tau=[0.1, 0.45])
  run(db, 0, pp, klasses=[MR], l=[.5], c=[0., 0.25, 0.5, 0.75, 1.], granularity=20)

  # run obama
  run(db, 11, pp, klasses=[NDT], c=[0., 0.25, 0.5, 0.75, 1.])
  run(db, 11, pp, klasses=[BDT], l=[.5], c=[0., 0.25, 0.5, 0.75, 1.], epsilon=[0.0001, 0.001], tau=[0.1, 0.45])
  run(db, 11, pp, klasses=[BDT], l=[.5], c=[1.], epsilon=[0.01, 0.1], tau=[0.1, 0.45])
  run(db, 11, pp, klasses=[MR], l=[.5], c=[0., 0.25, 0.5, 0.75, 1.], granularity=20)

  # run harddata 1
  run(db, 5, pp, klasses=[NDT], c=[0., 0.25, 0.5, 0.75, 1.])
  run(db, 5, pp, klasses=[BDT], l=[.5], c=[0., 0.25, 0.5, 0.75, 1.], epsilon=[0.0001, 0.001], tau=[0.1, 0.5])
  run(db, 5, pp, klasses=[BDT], l=[.5], c=[1.], epsilon=[0.01, 0.1])
  run(db, 5, pp, klasses=[MR], l=[.5], c=[0., 0.25, 0.5, 0.75, 1.], granularity=10)
  run(db, 5, pp, klasses=[MR], l=[.5], c=[0., 0.25, 0.5, 0.75, 1.], granularity=20)

  # run harddata 15  -- high dim
  run(db, 15, pp, klasses=[NDT], c=[0., 0.25, 0.5, 0.75, 1.])
  run(db, 15, pp, klasses=[BDT], l=[.5], c=[0., 0.25, 0.5, 0.75, 1.], epsilon=[0.0001, 0.001])
  run(db, 15, pp, klasses=[BDT], l=[.5], c=[1.], epsilon=[0.0001, 0.001, 0.01, 0.1])
  run(db, 15, pp, klasses=[MR], l=[.5], c=[0., 0.25, 0.5, 0.75, 1.], granularity=10)
  run(db, 15, pp, klasses=[MR], l=[.5], c=[0., 0.25, 0.5, 0.75, 1.], granularity=20)


  run(db, 0, pp, klasses=[NaiveMR], max_wait=20*60, c=[0., .25, .5, .75, 1.], granularity=20, naive=True)
  run(db, 11, pp, klasses=[NaiveMR], max_wait=20*60, c=[0., .25, .5, .75, 1.], granularity=20, naive=True)
  run(db, 5, pp, klasses=[NaiveMR], max_wait=20*60, c=[0., .25, .5, .75, 1.], granularity=20, naive=True)
  run(db, 15, pp, klasses=[NaiveMR], max_wait=20*60, c=[0., .25, .5, .75, 1.], granularity=20, naive=True)



  # run naive on intel_noon and increase the available columns
  # will take forever...
  run(db, 0, pp, klasses=[Naive], max_wait=30*60, cols=['voltage'], granularity=20, naive=True)
  run(db, 0, pp, klasses=[Naive], max_wait=30*60, cols=['voltage', 'humidity'], granularity=20, naive=True)
  run(db, 0, pp, klasses=[Naive], max_wait=30*60, cols=['voltage', 'humidity', 'light'], granularity=20, naive=True)
  run(db, 0, pp, klasses=[Naive], max_wait=30*60, cols=['moteid'], granularity=20, naive=True)
  run(db, 0, pp, klasses=[Naive], max_wait=30*60, granularity=20, naive=True)


  pp.close()



import json
import time
import pdb
import sys
import Orange
import orange
import heapq
sys.path.extend(['.', '..'])

from itertools import chain, ifilter


from ..learners.cn2sd.rule import fill_in_rules
from ..learners.cn2sd.refiner import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *
from ..util import *

from basic import Basic
from streamrangemerger import StreamRangeMerger
from rangemerger import RangeMerger, RangeMerger2
from merger import Merger
from grouper import Grouper, Blah

_logger = get_logger()
 


class MR(Basic):

  def __init__(self, *args, **kwargs):
    Basic.__init__(self, *args, **kwargs)
    self.best = []
    self.max_wait = kwargs.get('max_wait', 2 * 60 * 60) # 2 hours
    self.start = None
    self.stop = False
    self.n_rules_checked = 0
    self.naive = kwargs.get('naive', False)
    self.max_bests = 50
    self.max_complexity = kwargs.get('max_complexity', 3)

    self.checkpoints = []

    self.cost_clique = 0


  def __hash__(self):
    components = [
      self.__class__.__name__,
      str(self.aggerr.__class__.__name__),
      str(set(self.cols)),
      self.epsilon,
      self.tau,
      self.p,
      self.err_func.__class__.__name__,
      self.tablename,
      self.aggerr.keys,
      self.max_wait,
      self.c_range
    ]
    components = map(str, components)
    return hash('\n'.join(components))


  def setup_tables(self, full_table, bad_tables, good_tables, **kwargs):
    Basic.setup_tables(self, full_table, bad_tables, good_tables, **kwargs)
    self.grouper = Grouper(full_table, self) 

    self.SCORE_ID = add_meta_column(
            chain([full_table], bad_tables, good_tables),
            'SCOREVAR' 
    )



  def set_params(self, **kwargs):
    self.cols = kwargs.get('cols', self.cols)
    self.params.update(kwargs)
    self.good_thresh = 0.0001
    self.granularity = kwargs.get('granularity', self.granularity)

  def make_rules(self, cur_groups):
    if cur_groups == None:
      new_groups = self.grouper.initial_groups()
    else:
      new_groups = self.grouper.merge_groups(cur_groups)

    rules = {}

    for attrs, groups in new_groups:
      start = time.time()
      for ro in self.grouper(attrs, groups):
        if self.max_wait:
          self.n_rules_checked -= len(ro.rule.filter.conditions)
          if self.n_rules_checked <= 0:
            diff = time.time() - self.start
            if not self.checkpoints or diff - self.checkpoints[-1][0] > 10:
              if self.best:
                best_rule = max(self.best).rule
                self.checkpoints.append((diff, best_rule))
            self.stop = diff > self.max_wait
            self.n_rules_checked = 1000
          if self.stop:
            _logger.debug("wait %d > %d exceeded." % (diff, self.max_wait))
            return


        yield attrs, ro
#        print "group by\t%s\t%.4f" % (str([attr.name for attr in attrs]), time.time()-start)



  def __call__(self, full_table, bad_tables, good_tables, **kwargs):
    self.setup_tables(full_table, bad_tables, good_tables, **kwargs)
    self.update_status("running bottom up algorithm")
    for pairs in self.find_cliques():
      rules = [(b.rule, iteridx) for b, iteridx in pairs]
      yield rules
    self.update_status("bottom up algorithm done")


  def find_cliques(self):
    """
    table has been trimmed of extraneous columns.
    """
    #clusters = self.load_from_cache()
    #if clusters is not None:
      #yield clusters
      #return 

    rules = None
    self.best = []
    self.start = time.time()

    added = []
    nseen = 0
    niters = 0 
    while (niters < self.max_complexity and 
           not self.stop and 
           (rules is None or rules)):
      niters += 1
      self.update_status("running bottomup iter %d" % niters)
      _logger.debug("=========iter %d=========", niters)

      nadded = 0
      seen = set()
      nnewgroups = 0
      new_rules = defaultdict(list)
      
      # for each combination of attributes
      # prune the groups that are less influential than the parent group's 
      #  

      for attr, ro in self.make_rules(rules):
        nseen += 1
        if nseen % 50 == 0 and nseen > 0:
          self.update_status("bottomup processed %d rules" % nseen)

        if self.stop:
            break

        if self.top_k(ro):
          nadded += 1

        if self.naive:
            new_rules[attr] = [None]
            nnewgroups += 1
        elif self.prune_rule(ro):
            new_rules[attr].append(ro.group)
            nnewgroups += 1
        ro.rule.__examples__ = None


        if nadded % 25 == 0 and nadded > 0:
          newbests = filter(lambda c: c not in seen, self.best)
          seen.update(self.best)
          yield zip(newbests, [niters]*len(newbests))


      newbests = filter(lambda c: c not in seen, self.best)
      seen.update(self.best)
      yield zip(newbests, [niters]*len(newbests))
      if not nadded: 
        break 

      rules = new_rules
      if niters == 1:
        best = self.best
      else:
        best = set(self.best)
        if prev_best and prev_best in self.best:
          self.best.remove(prev_best)
        best = list(best)

      self.best = [max(self.best)] if self.best else []
      prev_best = max(self.best) if self.best else None

    _logger.debug("finished, merging now")
    self.cost_clique = time.time() - self.start

    #self.cache_results(clusters)

  def prune_rule(self, ro):
    if ro.npts < self.min_pts:
        _logger.debug("prune? %s\t%s", 'FALSE', str(ro))
        return False
    
    if (math.isnan(ro.bad_inf) or
        math.isnan(ro.good_inf) or
        math.isnan(ro.inf)):
        _logger.debug("prune? %s\t%s", 'FALSE', str(ro))
        return False
    

    # assuming the best case (the good_stat was zero)
    # would the influence beat the best so far across
    # the full c_range?
    if self.best:
      if ro.dominated_by(max(self.best)):
        _logger.debug("prune? %s\t%s", 'FALSE', str(ro))
        return False

    #if self.best and ro.best_inf <= max(self.best).inf:
    #    # if best tuple influence < rule influence:
    #    if ro.best_tuple_inf <= max(self.best).inf:
    #        _logger.debug("%s\t%s", 'FALSE', str(ro))
    #        return False

    # check max good influence
    if False and ro.good_inf < self.good_thresh:
        # TODO: can skip computing good_stats
        ro.good_skip = True


    #_logger.debug("%s\t%.4f\t%s", 'T', self.best and max(self.best).inf or 0, str(ro))
    return True


  def top_k(self, ro):
    n = 0
    best = self.best and max(self.best, key=lambda ro: ro.inf) or None
    if len(self.best) >= self.max_bests:
      bound = best.inf - self.best[0].inf
      thresh = self.best[0].inf + bound * 0.02
      if ro.inf <= thresh:
        return False

    if ro in self.best:
      return False
    if math.isnan(ro.inf):
      return False

    
    if len(self.best) < self.max_bests:
      n += 1            
      _logger.debug(str(ro))
      heapq.heappush(self.best, ro)
    else:
      n += 1            
      _logger.debug(str(ro))
      heapq.heapreplace(self.best, ro)
    
    best = best and max(best, ro) or ro

    return True

  @instrument
  def load_from_cache(self):
    import bsddb as bsddb3
    self.cache = bsddb3.hashopen('./dbwipes.mr.cache')
    try:
      myhash = str(hash(self))
      if myhash in self.cache and self.use_cache:
        self.update_status("loading partitions from cache")
        dicts, errors = json.loads(self.cache[myhash])
        clusters = map(Cluster.from_dict, dicts)
        for c in clusters:
          self.influence_cluster(c, self.full_table)
        return clusters
    except Exception as e:
      print e
      pdb.set_trace()
      pass
    finally:
      self.cache.close()
    return None


  @instrument
  def cache_results(self, clusters):
    import bsddb as bsddb3
    # save the clusters in a dictionary
    if self.use_cache:
      myhash = str(hash(self))
      self.cache = bsddb3.hashopen('./dbwipes.mr.cache')
      try:
        dicts = [c.to_dict() for c in clusters]
        errors = [c.error for c in clusters]
        self.cache[myhash] = json.dumps((dicts, errors))
      except Exception as e:
        print e
        pdb.set_trace()
        pass
      finally:
        self.cache.close()





import pdb
import datetime
import operator
import json
import logging
import orange
import Orange
import orngTree
import orngStat
import orngTest
import orngDisc
import Orange.feature as orf
import numpy as np


from datetime import datetime, date, timedelta
from datetime import time as dttime
from collections import deque, defaultdict
from dateutil.parser import parse as dateparse

from scorpionsql.db import *
from scorpionsql.sql import *
from scorpionsql.aggerror import *
from scorpionsql.sqlparser import *

from sharedobj import *
from util import *
from settings import *



_logger = get_logger()


# shared object
#
# get_provenance(results)
# for each result:
#   get provenance of result
#   rank provenance
#   SD2 cluster
# construct input dataset
# clean input dataset
# expand input dataset
# score(inputs)



def get_provenance(sharedobj, cols, keys):
    schema = sharedobj.rules_schema
    schema.update(cols)
    #_logger.debug( "get_provenance schema\t%s", schema)
    rows = sharedobj.get_tuples(keys, attrs=schema)
    if not len(rows):
        return None
    table = create_orange_table(rows, schema)
    return table


def get_provenance_split(sharedobj, cols, keys):
    """
    NOTE: #RNGE errors are when source table contains nulls.
    """
    gb = sharedobj.parsed.select.nonaggs[0]
    schema = sharedobj.rules_schema
    schema.update(cols)
    schema = list(schema)
    schema.append(str(gb))
    #_logger.debug( "get_provenance schema\t%s", schema)

    
    start = time.time()
    rows = sharedobj.get_tuples(keys, attrs=schema)
    print "dbquery cost %.4f" % (time.time() - start)
    if not len(rows): return None

    # create a column for the partitioning key so we can efficiently
    # apply as a mask
    cols = map(np.array, zip(*rows))

    # NOTE: we assume there are _no null values_
    #       enforce this by running misc/fixnulls.py
    keycol = np.array(cols.pop())
    schema.pop()

    start = time.time()
    # this domain still allows nulls in discrete attrs
    domain, funcs = orange_schema_funcs(cols, schema)
    print "schema funcs %.4f" % (time.time() - start)

    start = time.time()
    table = construct_orange_table(domain, schema, funcs, cols)
    domain = table.domain

    # removed nulls from domain NOTE: no more nulls
    print "create complete table %.4f" % (time.time() - start)

    start = time.time()
    tables = []
    data = table.to_numpyMA('ac')[0].data
    # find the rows that contain None values in discrete columns
    print "got numpy array %.4f" % (time.time() - start)

    start = time.time()
    for idx, key in enumerate(keys):
      # mask out rows that are not in this partition or have Nulls
      idxs = (keycol == key)
      partition_data = data[idxs,:]
      partition = Orange.data.Table(domain, partition_data)
      tables.append(partition)

    print "create orange time %.4f " % (time.time() - start)
    return tables






def extract_agg_vals(vals, col_type):
    fmts = [
      '%Y-%m-%dT%H:%M:%S.%fZ',
      '%Y-%m-%dT%H:%M:%S.%f',
      '%Y-%m-%dT%H:%M:%S',
      '%Y-%m-%dT%H:%M',
      '%Y-%m-%dT%H'
    ]
    for fmt in fmts:
      try:
        ret = [datetime.strptime(val, fmt) for val in vals]
        print vals
        if col_type == 'date':
          ret = [d.date() for d in ret]
        elif 'Z' in fmt:
          #ret = [d - timedelta(hours=5) for d in ret] # compensate for 'Z' +4 timezone
          pass
        return ret
      except Exception as e:
        pass

    try:
      ret = [datetime.strptime(val, '%Y-%m-%d').date() for val in vals]
      return ret
    except Exception as ee:
      print ee
      return vals




def is_discrete(attr, col):
  """
  Determines if a column is discrete or continuous.

  A huge number of special cases so we don't need to solve this 
  problem for the paper
  """
  if attr in [
    'epochid', 'voltage', 'xloc', 'yloc', 
    'est', 'height', 'width', 'atime', 'v',
    'light', 'humidity', 
    'finan_icu_days', 'dxage', 'job_count', 'b']:
      return False
  if attr in ['recipient_zip', 'sensor', 'moteid' 'file_num']:
      return True

  # continuous or discrete?
  # uniquecol = set(col)
  # nonnulls = filter(lambda x:x, col)
  # strtypes = map(lambda c: isinstance(c, str), nonnulls[:20])
  # istypestr = reduce(operator.or_, strtypes) if strtypes else True

  # if its strings
  nonnulls = filter(lambda x:x, col)
  strtypes = map(lambda c: isinstance(c, basestring), nonnulls[:20])
  istypestr = reduce(operator.or_, strtypes) if strtypes else True
  if istypestr:
      return True

  # if its floats
  try:
      isints = map(lambda v: int(v) == float(v), nonnulls)
      istypeint = reduce(operator.and_, isints) if isints else True
      if not istypeint:
          return False
  except:
      pass

  # or if there are too many unique values
  uniquecols = set(nonnulls)
  if len(uniquecols) > 0.05 * len(col):
      return False

  return True

def detect_discrete_cols(rows, attrs):
  attrs = list(attrs)
  cols = map(list, zip(*rows))
  dcols = []
  for idx, (attr, col) in enumerate(zip(attrs, cols)):
    if is_discrete(attr, col):
      dcols.append(attr)
  return dcols


def orange_schema_funcs(cols, attrs):
  """
  Args
    cols:  values of table in columnar format
           list of columns. each column is a list of values
    attrs: list of attribute names for each column
  @return (Domain, [funcs]) funcs is list of functions to apply to 
          the corresponding column value
  """
  # orngDisc.orngDisc.entropyDiscretization('')
  features = []
  funcs = []
  for idx, (attr, col) in enumerate(zip(attrs, cols)):
    start = time.time()
    bdiscrete = is_discrete(attr, col)

    if bdiscrete:
      feature = Orange.feature.Discrete(attr, values=map(str, set(col)))
      func = str
    else:
      feature = Orange.feature.Continuous(attr)
      func = lambda v: v is not None and float(v) or 0.

    print "\tschema func for %s\t%s\t%.4f" % (attr, func, time.time() - start)

    features.append(feature)
    funcs.append(func)

  domain = Orange.data.Domain(features)
  return domain, funcs

def domain_rm_nones(domain):
  features = []
  for attr in domain:
    if isinstance(attr, Orange.feature.Discrete):
      vals = attr.values
      vals = filter(lambda v: v not in [None, 'None'], vals)
      attr = Orange.feature.Discrete(attr.name, values=vals)
    features.append(attr)
  return Orange.data.Domain(features)


def construct_orange_table(domain, attrs, funcs, cols):
  """
  Args:
    cols: list of numpy arrays
  """
  start = time.time()
  for idx in xrange(len(cols)):
    if funcs[idx] == str:
      print cols[idx]
      cols[idx] = cols[idx].astype(str)
    else:
      try:
        f = np.frompyfunc(funcs[idx], 1, 1)
        cols[idx] = funcs[idx](cols[idx])
      except Exception as e:
        _logger.debug(str(e))
        f = np.vectorize(funcs[idx])
        try:
          cols[idx] = f(cols[idx])
        except Exception as e:
          _logger.debug(str(e))
          cols[idx] = map(funcs[idx], cols[idx])

  print "func(cols) %.4f" % (time.time() - start)

  start = time.time()
  rows = map(list, zip(*cols))
  print "transpose %.4f" % (time.time() - start)
  start = time.time()
  table = Orange.data.Table(domain)
  table.extend(rows)
  print "created table %.4f" % (time.time() - start)
  return table


def create_orange_table(rows, attrs):
  cols = map(np.array, zip(*rows))
  domain, funcs = orange_schema_funcs(cols, attrs)
  attrs = list(attrs)
  return construct_orange_table(domain, attrs, funcs, cols)





import os
import time
import json
import logging
import sys
import random
import math
import matplotlib
import numpy as np

from collections import defaultdict
from itertools import *
from Orange.data.filter import *
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt

nan = float('nan')
inf = float('inf')

matplotlib.use("Agg")

"""
Super generic functions that can be used across modules
Global values that are used in every module
"""
LINESTYLES = ['-', '--o']#, '--', '-.', ',', 'o', 'v', '^', '>', '1', '*', ':']
COLORS = ['#d62728', '#1f77b4', '#9467bd', '#ff7f0e', '#2ca02c',
          '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf',
          '#000000', 'red', 'orange', 'blue', 'purple', 'grey',
          'green', 'yellow', 'purple']
#COLORS = ['%s%s' % (color, m) for color in __colors__ for m in __markers__]
NAN = float('nan')

class GlobalLogger(object):
    def __init__(self):
        self._logger = None
    
    def __call__(self, fname='/tmp/scorpion.log', flevel=logging.DEBUG, plevel=logging.WARNING):
        if self._logger:
            # set levels
            return self._logger

        self._logger = logging.getLogger()
        self._logger.setLevel(logging.DEBUG)

        #formatter = logging.Formatter('%(asctime)s - %(lineno)s - %(levelname)s - %(message)s')
        formatter = logging.Formatter('%(lineno)s - %(levelname)s - %(message)s')

        fname = os.path.abspath(fname)
        fh = logging.FileHandler(fname)
        fh.setLevel(flevel)
        fh.setFormatter(formatter)
        self._logger.addHandler(fh)

        ph = logging.StreamHandler(sys.stdout)
        ph.setLevel(plevel)
        ph.setFormatter(formatter)
        self._logger.addHandler(ph)
        return self._logger

get_logger = GlobalLogger()


def instrument(fn):
  func_name = fn.__name__
  nargs = fn.func_code.co_argcount
  def w(self, *args, **kwargs):
    start = time.time()
    if len(args) == nargs:
      ret = fn(*args, **kwargs)
    else:
      ret = fn(self, *args, **kwargs)

    if hasattr(self, 'stats'):
      if func_name not in self.stats:
          self.stats[func_name] = [0, 0]
      self.stats[func_name][0] += (time.time() - start)
      self.stats[func_name][1] += 1
    return ret
  return w


# JSON Encoder
class ScorpionEncoder(json.JSONEncoder):
  def default(self, o):
    if isinstance(o, float):
      if o == float('inf'):
        return 1e100
      elif o == float('-inf'):
        return -1e100

    if hasattr(o, 'isoformat'):
      s =  o.isoformat()
      if not s.endswith("Z"):
        s += 'Z'
      return s
    return super(ScorpionEncoder, self).default(o)



def block_iter(l, nblocks=2):
    """
    partitions l into nblocks blocks and returns generator over each block
    @param l list
    @param nblocks number of blocks to partition l into
    """
    blocksize = int(math.ceil(len(l) / float(nblocks)))
    i = 0
    while i < len(l):
        yield l[i:i+blocksize]
        i += blocksize      
    

def rm_dups(seq, key=lambda e:e):
    """
    returns a list of items with duplicates removed
    @note from http://www.peterbe.com/plog/uniqifiers-benchmark
    """
    seen = set()
    seen_add = seen.add
    return [ x for x in seq if key(x) not in seen and not seen_add(key(x))]


def groupby(iterable, key=lambda: 1):
  ret = defaultdict(list)
  for i in iterable:
    ret[key(i)].append(i)
  return ret.iteritems()



def powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))

def valid_number(v):
  return v is not None and v != -inf and not(math.isnan(v)) and v != inf


class LinearFit(object):

    def __init__(self, data=None):
        self.x2 = 0.
        self.y2 = 0.
        self.xm = 0.
        self.ym = 0.
        self.xy = 0.
        self.n = 0.
        self.ssxx = 0.
        self.ssyy = 0.
        self.ssxy = 0.
        self.b = 0.
        self.a = 0.
        self.r2 = 0.
        self.p = 0.
        if data is not None:
            self(data)

    def __call__(self, xs, ys):
        #pts = data
        #xs, ys = pts[:,0], pts[:,1]
        n = len(xs)
        self.x2 = np.dot(xs, xs)
        self.y2 = np.dot(ys, ys)
        self.xm = np.mean(xs)
        self.ym = np.mean(ys)
        self.xy = np.dot(xs, ys)
        self.n  = n

        self.ssxx = self.x2 - self.n*(self.xm**2)
        self.ssyy = self.y2 - self.n*self.ym*self.ym
        self.ssxy = self.xy - self.n*self.xm*self.ym

        self.b = self.ssxy / self.ssxx # slope
        self.a = self.ym - self.b*self.xm  # intercept
        self.r2 = self.ssxy**2 / (self.ssxx*self.ssyy)   # correlation coefficient
        self.p = self.ssxy / math.sqrt(self.ssxx*self.ssyy)  # pearson's = sqrt(r2)
        return self.a, self.b, self.r2, self.p


    def delta(self, add=None, rm=None):
        if add is None and rm is None:
            return self.a, self.b, self.r2, self.p
        
        x2, y2 = self.x2, self.y2
        xm, ym, xy = self.xm, self.ym, self.xy
        n = self.n

        if add is not None:
            #xs, ys = add[:,0], add[:,1]
            xs, ys = tuple(add)
            if n == len(xs):
                return 0,0,0,0
            
            x2 += np.dot(xs, xs)
            y2 += np.dot(ys, ys)
            xm = (xm * n + xs.sum()) / (n + len(xs))
            ym = (ym * n + ys.sum()) / (n + len(ys))
            xy += np.dot(xs, ys)
            n += len(xs)
        if rm is not None:
            #xs, ys = rm[:,0], rm[:,1]
            xs, ys = tuple(rm)
            if n == len(xs):
                return 0,0,0,0
            
            x2 -= np.dot(xs, xs)
            y2 -= np.dot(ys, ys)
            xm = (xm * n - xs.sum()) / (n - len(xs))
            ym = (ym * n - ys.sum()) / (n - len(ys))
            xy -= np.dot(xs, ys)
            n -= len(xs)

       
        ssxx = x2 - (n*xm*xm)
        ssyy = y2 - (n*ym*ym)
        ssxy = xy - (n*xm*ym)

        if n == 0 or ssxx <= 0 or ssyy <= 0:
            return NAN, NAN, NAN, NAN
        
        b = ssxy / ssxx # slope
        a = ym - b*xm   # intercept
        r2 = ssxy**2 / (ssxx*ssyy)   # correlation coefficient
        p = ssxy / math.sqrt(ssxx*ssyy)  # pearson's = sqrt(r2)
        return a, b, r2, p


    def linear(self, add=None, rm=None):
        a,b,r2,p = self.delta(add=add, rm=rm)
        return a, b, r2

    def corr(self, add=None, rm=None):
        a,b,r2,p = self.delta(add=add, rm=rm)
        return p

    def pearson(self, **kwargs):
        return self.corr(**kwargs)

if __name__ == '__main__':
    


    
    import random
    random.seed(0)
    pts = []
    for i in xrange(1000):
        pts.append( (i, i + random.random()) )
    pts = np.array(pts)

    
    lf = LinearFit()
    print lf.corr(pts)
    print lf.corr(pts)    




import json
from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool

class Status(object):
  def __init__(self, reqid=None):
    self.engine = create_engine("postgresql://localhost/status", poolclass=NullPool)
    self.db = self.engine.connect()
    self.create_table()
    self.reqid = reqid
    self.reqid = self.get_req_id()

  def close(self):
    try:
      self.db.close()
    except:
      pass
    try:
      self.engine.dispose()
    except:
      pass

  def create_table(self):
    stmts = [
      "create table requests(id serial)",
      "create table status(id serial, reqid int, status varchar)",
      "create table rules(reqid int, label text, val text)"
    ]
    for stmt in stmts:
      try:
        self.db.execute(stmt)
      except:
        pass

  def cleanup(self):
    self.db.execute("delete from status where reqid = %s", self.reqid)
    self.db.execute("delete from requests where id = %s", self.reqid)
    self.reqid = None

  def get_req_id(self):
    if self.reqid is not None: 
      q = "select count(*) from requests where id = %s"
      count = self.db.execute(q, self.reqid).fetchone()[0] 
      if count > 0: 
        return self.reqid
    q = "insert into requests values (default) returning id"
    return self.db.execute(q).fetchone()[0]

  def update_status(self, s):
    print "update status", self.reqid, s
    q = "insert into status values (default, %s, %s)"
    self.db.execute(q, self.reqid, s)

  def latest_status(self):
    q = "select status from status where reqid = %s order by id desc limit 1"
    rows = self.db.execute(q, self.reqid).fetchall()
    if len(rows):
      return rows[0][0]
    return 'no status yet'

  def update_rules(self, label, rules):
    q = """update rules set val=%s where reqid=%s and label = %s;
    insert into rules
      select %s, %s, %s 
      where not exists (
        select 1 from rules where reqid=%s and label = %s
      );
    """
    from scorpion.learners.cn2sd.rule import rule_to_json
    jsons = []
    for rule in rules:
      if isinstance(rule, dict):
        jsons.append(rule)
      else:
        jsons.append(rule_to_json(rule, yalias=label))
    val = json.dumps(jsons)
    self.db.execute(q, 
        val, self.reqid, label, 
        self.reqid, label, val, 
        self.reqid, label)

  def get_rules(self):
    q = "select label, val from rules where reqid = %s"
    rows = self.db.execute(q, self.reqid).fetchall()
    if len(rows):
      ret = [ (str(row[0]), json.loads(row[1])) for row in rows ]
      return ret
    return []
    
    




from scorpion.sharedobj import *
from scorpion.sql import *
from scorpion.sqlparser import *
from scorpion.db import *



parsed = parse_sql("select hr, avg(temp) from readings group by hr")
parsed.where.append("sensor in %s")
params = [('1', '2', '18')]

db = connect("intel")

obj = SharedObj(db, parsed=parsed, params=params)
for t in obj.get_filter_rows(['2004-03-05 12:00:00']):
  print t
